import inspect
import typing
from typing import Any, Dict, List, Mapping, Optional, Sequence, Type, Union

from apischema.json_schema import (
    JsonSchemaVersion,
    definitions_schema,
    deserialization_schema,
    serialization_schema,
)
from squall import routing
from squall.datastructures import DefaultPlaceholder
from squall.openapi.constants import (
    METHODS_WITH_BODY,
    REF_PREFIX,
    STATUS_CODES_WITH_NO_BODY,
)
from squall.params import Body
from squall.responses import Response
from squall.routing import APIRoute, WebSocketRoute
from squall.routing_.utils import HeadParam
from squall.utils import generate_operation_id_for_path

AnyRoute = Union[APIRoute, WebSocketRoute]


validation_error_definition = {
    "title": "ValidationError",
    "type": "object",
    "properties": {
        "loc": {"title": "Location", "type": "array", "items": {"type": "string"}},
        "msg": {"title": "Message", "type": "string"},
        "type": {"title": "Error Type", "type": "string"},
    },
    "required": ["loc", "msg", "type"],
}

validation_error_response_definition = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "detail": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": REF_PREFIX + "ValidationError"},
        }
    },
}


param_validation_error_response_definition = {
    "title": "HTTPBadRequestError",
    "type": "object",
    "properties": {
        "details": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": REF_PREFIX + "ValidationError"},
        }
    },
}

status_code_ranges: Dict[str, str] = {
    "1XX": "Information",
    "2XX": "Success",
    "3XX": "Redirection",
    "4XX": "Client Error",
    "5XX": "Server Error",
    "DEFAULT": "Default Response",
}


type_mapping: Dict[str, str] = {
    "int": "integer",
    "Decimal": "number",
    "float": "number",
    "bool": "boolean",
    "str": "string",
    "bytes": "string",
}


def generate_operation_id(*, route: routing.APIRoute, method: str) -> str:
    if route.operation_id:
        return route.operation_id
    path: str = route.path_format
    return generate_operation_id_for_path(name=route.name, path=path, method=method)


def generate_operation_summary(*, route: routing.APIRoute, method: str) -> str:
    if route.summary:
        return route.summary
    return route.name.replace("_", " ").title()


def get_openapi_operation_metadata(
    *, route: routing.APIRoute, method: str
) -> Dict[str, Any]:
    operation: Dict[str, Any] = {}
    if route.tags:
        operation["tags"] = route.tags
    operation["summary"] = generate_operation_summary(route=route, method=method)
    if route.description:
        operation["description"] = route.description
    operation["operationId"] = generate_operation_id(route=route, method=method)
    if route.deprecated:
        operation["deprecated"] = route.deprecated
    return operation


class OpenAPIRoute:
    def __init__(self, route: APIRoute, version: JsonSchemaVersion) -> None:
        self.version = version
        self.route = route
        self._response_class = route.response_class
        self.request_schemas: typing.Set[Any] = set()
        self.response_schemas: typing.Set[Any] = set()

    @property
    def response_class(self) -> Type[Response]:
        if isinstance(self._response_class, DefaultPlaceholder):
            current_response_class: Type[Response] = self._response_class.value
        else:
            current_response_class = self._response_class
        return current_response_class

    @property
    def default_status_code(self) -> str:
        code = str(200)
        if self.route.status_code is not None:
            code = str(self.route.status_code)
        else:
            response_signature = inspect.signature(self.response_class.__init__)
            status_code_param = response_signature.parameters.get("status_code")
            if status_code_param is not None:
                code = str(status_code_param.default)

        return code

    @property
    def responses(self) -> Dict[Union[str, int], Any]:
        responses: Dict[Union[int, str], Any] = {}
        # Default response
        status_code: Union[str, int] = self.default_status_code
        responses[status_code] = {"description": self.route.response_description}
        if (
            self.response_class.media_type
            and status_code not in STATUS_CODES_WITH_NO_BODY
        ):
            response_schema: Mapping[str, Any] = {"type": "string"}
            if (
                self.response_class.media_type is not None
                and self.response_class.media_type[-4:] == "json"
            ):
                if self.route.response_field:
                    response_schema = serialization_schema(
                        self.route.response_field.model,
                        all_refs=True,
                        version=self.version,
                    )
                    self.response_schemas.add(self.route.response_field.model)
                else:
                    response_schema = {}
            media_type = self.response_class.media_type

            if media_type is None and response_schema:
                media_type = "application/json"
            responses[status_code]["content"] = {
                media_type: {"schema": response_schema}
            }

        # Errors
        if self.route.head_params:
            responses["400"] = {
                "description": "Parameters Validation Error",
                "content": {
                    "application/json": {
                        "schema": {"$ref": REF_PREFIX + "HTTPBadRequestError"}
                    }
                },
            }
        # For request_field
        if self.route.request_field is not None:
            responses["422"] = {
                "description": "Request Body Validation Error",
                "content": {
                    "application/json": {
                        "schema": {"$ref": REF_PREFIX + "HTTPValidationError"}
                    }
                },
            }

        # User defined responses
        for status_code, response in self.route.responses.items():
            if status_code not in status_code_ranges and status_code != "default":
                status_code = str(int(status_code))

            if status_code not in responses:
                responses[status_code] = {}

            if model := response.get("model"):
                response_schema = serialization_schema(
                    model, all_refs=True, version=self.version
                )
                media_type = self.response_class.media_type or "application/json"

                responses[status_code]["content"] = {
                    media_type: {"schema": response_schema}
                }
                self.response_schemas.add(model)
            if description := response.get("description"):
                responses[status_code]["description"] = description

            if content := response.get("content"):
                if "content" not in responses[status_code]:
                    responses[status_code]["content"] = content.copy()
                else:
                    responses[status_code]["content"].update(content)

        return responses

    @property
    def request_body(self) -> Optional[Dict[str, Any]]:
        if not self.route.request_field:
            return None

        self.request_schemas.add(self.route.request_field.model)
        response_schema = deserialization_schema(
            self.route.request_field.model,
            all_refs=True,
            version=self.version,
        )

        settings: Optional[Body] = self.route.request_field.settings

        media_type = getattr(settings, "media_type", "application/json")
        result = {"required": getattr(settings, "required", True)}
        content: Dict[str, Dict[str, Any]] = {media_type: {}}

        if settings:
            if settings.examples:
                content[media_type]["examples"] = settings.examples
            elif settings.example:
                content[media_type]["example"] = settings.example

        content[media_type]["schema"] = response_schema
        result["content"] = content
        return result

    @staticmethod
    def get_param_spec(param: HeadParam) -> Dict[str, Any]:
        source_mapping = {
            "path_params": "path",
            "query_params": "query",
            "headers": "header",
            "cookies": "cookie",
        }

        item: Dict[str, Any] = {}
        item["type"] = type_mapping.get(param.convertor, "string")

        schema: Dict[str, Any]
        if param.is_array:
            schema = {"type": "array", "items": item}
        else:
            schema = item
            if param.statements.get("ge") is not None:
                schema["minimum"] = param.statements["ge"]
                schema["exclusiveMinimum"] = False
            elif param.statements.get("gt") is not None:
                schema["minimum"] = param.statements["gt"]
                schema["exclusiveMinimum"] = True

            if param.statements.get("le") is not None:
                schema["maximum"] = param.statements["le"]
                schema["exclusiveMaximum"] = False
            elif param.statements.get("lt") is not None:
                schema["maximum"] = param.statements["lt"]
                schema["exclusiveMaximum"] = True

        result = {
            "required": param.default == Ellipsis,
            "schema": schema,
            "name": param.alias or param.name,
            "in": source_mapping[param.source],
        }

        description = getattr(param._default, "description", None)
        if description is not None:
            result["description"] = description

        example = getattr(param._default, "example", None)
        if example is not None:
            result["example"] = example

        examples = getattr(param._default, "examples", None)
        if examples:
            result["examples"] = examples

        deprecated = getattr(param._default, "deprecated", None)
        if deprecated:
            result["deprecated"] = True

        return result

    @property
    def head_params(self) -> List[Dict[str, Any]]:
        parameters: List[Dict[str, Any]] = []
        for param in self.route.head_params:
            parameters.append(self.get_param_spec(param))
        return parameters

    @property
    def spec(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {}
        if self.route.methods is None:
            return data

        for method in self.route.methods:
            operation = get_openapi_operation_metadata(route=self.route, method=method)
            if parameters := self.head_params:
                operation["parameters"] = parameters
            operation["responses"] = self.responses

            if self.route.openapi_extra is not None:
                operation.update(self.route.openapi_extra)

            if method in METHODS_WITH_BODY:
                if request_body := self.request_body:
                    operation["requestBody"] = request_body

            data[method.lower()] = operation

        return data


def get_openapi(
    *,
    title: str,
    version: str,
    openapi_version: str = "3.0.2",
    description: Optional[str] = None,
    routes: Sequence[AnyRoute],
    tags: Optional[List[Dict[str, Any]]] = None,
    servers: Optional[List[Dict[str, Union[str, Any]]]] = None,
    terms_of_service: Optional[str] = None,
    contact: Optional[Dict[str, Union[str, Any]]] = None,
    license_info: Optional[Dict[str, Union[str, Any]]] = None,
) -> Dict[str, Any]:
    info: Dict[str, Any] = {"title": title, "version": version}
    if description:
        info["description"] = description
    if terms_of_service:
        info["termsOfService"] = terms_of_service
    if contact:
        info["contact"] = contact
    if license_info:
        info["license"] = license_info
    output: Dict[str, Any] = {"openapi": openapi_version, "info": info}
    if servers:
        output["servers"] = servers
    components: Dict[str, Any] = {}
    paths: Dict[str, Dict[str, Any]] = {}

    _version = JsonSchemaVersion.OPEN_API_3_1
    request_schemas = set()
    response_schemas = set()
    for route in routes:
        if isinstance(route, WebSocketRoute):
            continue

        if not route.include_in_schema or not route.path_format:
            continue

        openapi_route = OpenAPIRoute(route, version=_version)
        if route.path_format in paths:
            paths[route.path_format].update(openapi_route.spec)
        else:
            paths[route.path_format] = openapi_route.spec
        response_schemas.update(openapi_route.response_schemas)
        request_schemas.update(openapi_route.request_schemas)

    schemas = set.union(request_schemas, response_schemas)
    if schemas:
        components["schemas"] = definitions_schema(
            deserialization=list(schemas),
            all_refs=True,
            version=_version,
        )

        # Should be removed from here. Schema should be defined via handlers
        if request_schemas:
            if "ValidationError" not in components["schemas"]:
                components["schemas"]["ValidationError"] = validation_error_definition
            if "HTTPValidationError" not in components["schemas"]:
                components["schemas"][
                    "HTTPValidationError"
                ] = validation_error_response_definition
        if "HTTPBadRequestError" not in components["schemas"]:
            components["schemas"][
                "HTTPBadRequestError"
            ] = param_validation_error_response_definition

    if components:
        output["components"] = components
    output["paths"] = paths
    if tags:
        output["tags"] = tags
    return output



from squall import Squall
from squall.testclient import TestClient

app = Squall(
    servers=[
        {"url": "/", "description": "Default, relative server"},
        {
            "url": "http://staging.localhost.mtag.dev:8000",
            "description": "Staging but actually localhost still",
        },
        {"url": "https://prod.example.com"},
    ]
)


@app.get("/foo")
def foo():
    return {"message": "Hello World"}


client = TestClient(app)


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "servers": [
        {"url": "/", "description": "Default, relative server"},
        {
            "url": "http://staging.localhost.mtag.dev:8000",
            "description": "Staging but actually localhost still",
        },
        {"url": "https://prod.example.com"},
    ],
    "paths": {
        "/foo": {
            "get": {
                "summary": "Foo",
                "operationId": "foo_foo_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        }
    },
}


def test_openapi_servers():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_app():
    response = client.get("/foo")
    assert response.status_code == 200, response.text



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['squall',
 'squall.middleware',
 'squall.openapi',
 'squall.routing_',
 'squall.validators']

package_data = \
{'': ['*']}

install_requires = \
['starlette ==0.16.0',
 'orjson >=3.6.4,<4.0.0',
 'apischema >=0.16.2,<1.0.0',
 'isal >=0.11.1,<1.0.0']

extras_require = \
{'all': ['requests >=2.24.0,<3.0.0',
         'jinja2 >=2.11.2,<4.0.0',
         'python-multipart >=0.0.5,<0.0.6',
         'itsdangerous >=1.1.0,<3.0.0',
         'pyyaml >=5.3.1,<6.0.0',
         'orjson >=3.6.4,<4.0.0',
         'email_validator >=1.1.1,<2.0.0',
         'uvicorn[standard] >=0.12.0,<0.16.0'],
 'dev': ['python-jose[cryptography] >=3.3.0,<4.0.0',
         'passlib[bcrypt] >=1.7.2,<2.0.0',
         'autoflake >=1.4.0,<2.0.0',
         'flake8 >=3.8.3,<4.0.0',
         'uvicorn[standard] >=0.12.0,<0.16.0'],
 'doc': ['mkdocs >=1.1.2,<2.0.0',
         'mkdocs-material >=7.1.9,<8.0.0',
         'mdx-include >=1.4.1,<2.0.0',
         'mkdocs-markdownextradata-plugin >=0.1.7,<0.3.0',
         'typer-cli >=0.0.12,<0.0.13',
         'pyyaml >=5.3.1,<6.0.0',
         'pydantic >=1.8.2,<2.0.0'],
 'test': ['pytest >=6.2.4,<7.0.0',
          'pytest-cov >=2.12.0,<4.0.0',
          'pytest-mock >=3.6.1,<4.0.0',
          'mypy ==0.910',
          'flake8 >=3.8.3,<4.0.0',
          'black ==21.9b0',
          'isort >=5.0.6,<6.0.0',
          'requests >=2.24.0,<3.0.0',
          'httpx >=0.14.0,<0.19.0',
          'email_validator >=1.1.1,<2.0.0',
          'sqlalchemy >=1.3.18,<1.5.0',
          'peewee >=3.13.3,<4.0.0',
          'databases[sqlite] >=0.3.2,<0.6.0',
          'python-multipart >=0.0.5,<0.0.6',
          'flask >=1.1.2,<3.0.0',
          'anyio[trio] >=3.2.1,<4.0.0',
          'types-orjson ==3.6.0'],
 "test:python_version<'3.7'": ['types-dataclasses ==0.1.7']}

setup(name='python-squall',
      version='0.2.1',
      description='Squall ASGI REST framework',
      author=None,
      author_email='Stanislav Dubrovskyi <s.dubrovskyi@cleverdec.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.9.0',
     )



from dataclasses import dataclass, field

from squall import Squall
from squall.testclient import TestClient

app = Squall()


@dataclass
class Model:
    pass


@dataclass
class Model2:
    a: Model = field()


@dataclass
class Model3:
    c: Model = field()
    d: Model2 = field()


@app.get("/", response_model=Model3)
def f():
    return {"c": {}, "d": {"a": {}}}


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "Model3": {
                "type": "object",
                "properties": {
                    "c": {"$ref": "#/components/schemas/Model"},
                    "d": {"$ref": "#/components/schemas/Model2"},
                },
                "required": ["c", "d"],
                "additionalProperties": False,
            },
            "Model": {"type": "object", "additionalProperties": False},
            "Model2": {
                "type": "object",
                "properties": {"a": {"$ref": "#/components/schemas/Model"}},
                "required": ["a"],
                "additionalProperties": False,
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/": {
            "get": {
                "summary": "F",
                "operationId": "f__get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/Model3"}
                            }
                        },
                    }
                },
            }
        }
    },
}


client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema


def test_get_api_route():
    response = client.get("/")
    assert response.status_code == 200, response.text
    assert response.json() == {"c": {}, "d": {"a": {}}}



import typing
from dataclasses import dataclass

from squall import Response, Squall
from squall.responses import JSONResponse
from squall.testclient import TestClient

app = Squall()


class JsonApiResponse(JSONResponse):
    media_type = "application/vnd.api+json"


@dataclass
class Error:
    status: str
    title: str


@dataclass
class JsonApiError:
    errors: typing.List[Error]


@app.get(
    "/a",
    response_class=Response,
    responses={500: {"description": "Error", "model": JsonApiError}},
)
async def a():
    pass  # pragma: no cover


@app.get("/b", responses={500: {"description": "Error", "model": Error}})
async def b():
    pass  # pragma: no cover


openapi_schema = {
    "openapi": "3.0.2",
    "info": {"title": "Squall", "version": "0.1.0"},
    "components": {
        "schemas": {
            "JsonApiError": {
                "type": "object",
                "properties": {
                    "errors": {
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/Error"},
                    }
                },
                "required": ["errors"],
                "additionalProperties": False,
            },
            "Error": {
                "type": "object",
                "properties": {
                    "status": {"type": "string"},
                    "title": {"type": "string"},
                },
                "required": ["status", "title"],
                "additionalProperties": False,
            },
            "HTTPBadRequestError": {
                "title": "HTTPBadRequestError",
                "type": "object",
                "properties": {
                    "details": {
                        "title": "Detail",
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/ValidationError"},
                    }
                },
            },
        }
    },
    "paths": {
        "/a": {
            "get": {
                "summary": "A",
                "operationId": "a_a_get",
                "responses": {
                    "200": {"description": "Successful Response"},
                    "500": {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/JsonApiError"}
                            }
                        },
                        "description": "Error",
                    },
                },
            }
        },
        "/b": {
            "get": {
                "summary": "B",
                "operationId": "b_b_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    },
                    "500": {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": "#/components/schemas/Error"}
                            }
                        },
                        "description": "Error",
                    },
                },
            }
        },
    },
}


client = TestClient(app)


def test_openapi_schema():
    response = client.get("/openapi.json")
    assert response.status_code == 200, response.text
    assert response.json() == openapi_schema



import pytest
from squall import Squall
from squall.compression import Compression, GzipBackend, ZlibBackend
from squall.responses import HTMLResponse, JSONResponse, PlainTextResponse
from squall.testclient import TestClient


@pytest.mark.parametrize(
    "backends,content_encoding",
    [
        [[GzipBackend(), ZlibBackend()], "gzip"],
        [[ZlibBackend()], "deflate"],
        [[GzipBackend()], "gzip"],
    ],
)
def test_compression_backends(backends, content_encoding):
    app = Squall(compression=Compression(backends=backends, minimal_size=1))

    @app.get("/")
    def get_root():
        return {"msg": "Hello World"}

    client = TestClient(app)
    response = client.get("/")
    assert response.json() == {"msg": "Hello World"}
    assert response.headers["Content-Encoding"] == content_encoding


@pytest.mark.parametrize(
    "response_class,response,expected_response",
    [
        [HTMLResponse, "Hello World", b"Hello World"],
        [JSONResponse, {"msg": "Hello World"}, b'{"msg":"Hello World"}'],
        [PlainTextResponse, "Hello World", b"Hello World"],
    ],
)
def test_compression_response_classes(response_class, response, expected_response):
    app = Squall(compression=Compression(minimal_size=1))

    @app.get("/", response_class=response_class)
    def get_root():
        return response

    client = TestClient(app)
    response = client.get("/")
    assert response.content == expected_response



from typing import Any, Callable, Iterable, Type, TypeVar

from starlette.datastructures import URL as URL  # noqa: F401
from starlette.datastructures import Address as Address  # noqa: F401
from starlette.datastructures import FormData as FormData  # noqa: F401
from starlette.datastructures import Headers as Headers  # noqa: F401
from starlette.datastructures import QueryParams as QueryParams  # noqa: F401
from starlette.datastructures import State as State  # noqa: F401
from starlette.datastructures import UploadFile as StarletteUploadFile
from starlette.datastructures import URLPath as URLPath  # noqa: F401


class UploadFile(StarletteUploadFile):
    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        yield cls.validate

    @classmethod
    def validate(cls: Type["UploadFile"], v: Any) -> Any:
        if not isinstance(v, StarletteUploadFile):
            raise ValueError(f"Expected UploadFile, received: {type(v)}")
        return v


class DefaultPlaceholder:
    """
    You shouldn't use this class directly.

    It's used internally to recognize when a default value has been overwritten, even
    if the overridden default value was truthy.
    """

    def __init__(self, value: Any):
        self.value = value

    def __bool__(self) -> bool:
        return bool(self.value)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, DefaultPlaceholder) and o.value == self.value


DefaultType = TypeVar("DefaultType")


def Default(value: DefaultType) -> DefaultType:
    """
    You shouldn't use this function directly.

    It's used internally to recognize when a default value has been overwritten, even
    if the overridden default value was truthy.
    """
    return DefaultPlaceholder(value)  # type: ignore



from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Any, Dict, Optional, Union

Number = Union[int, float, Decimal]


class ValidatorTypes(Enum):
    numeric = "numeric"
    string = "string"


@dataclass
class Num:
    in_ = ValidatorTypes.numeric
    gt: Optional[Number] = None
    ge: Optional[Number] = None
    lt: Optional[Number] = None
    le: Optional[Number] = None


@dataclass
class Str:
    in_ = ValidatorTypes.string
    min_len: Optional[int] = None
    max_len: Optional[int] = None


class ParamTypes(Enum):
    query = "query_params"
    header = "headers"
    path = "path_params"
    cookie = "cookies"


class ParamOpenAPI(Enum):
    query = "query"
    header = "header"
    path = "path"
    cookie = "cookie"


@dataclass
class CommonParam:
    default: Any = Ellipsis
    alias: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    valid: Optional[Union[Str, Num]] = None
    example: Optional[Any] = None
    examples: Optional[Dict[str, Any]] = None
    deprecated: Optional[bool] = None


@dataclass
class Path(CommonParam):
    in_ = ParamTypes.path
    openapi_ = ParamOpenAPI.path


@dataclass
class Query(CommonParam):
    in_ = ParamTypes.query
    openapi_ = ParamOpenAPI.query


@dataclass
class Header(CommonParam):
    in_ = ParamTypes.header
    openapi_ = ParamOpenAPI.header


@dataclass
class Cookie(CommonParam):
    in_ = ParamTypes.cookie
    openapi_ = ParamOpenAPI.cookie


@dataclass
class Body:
    default: Any = Ellipsis
    media_type: str = "application/json"
    embed: bool = False
    title: Optional[str] = None
    description: Optional[str] = None
    required: Optional[bool] = None
    example: Optional[Any] = None
    examples: Optional[Dict[str, Any]] = None


@dataclass
class Form(Body):
    media_type: str = "application/x-www-form-urlencoded"
    alias: Optional[str] = None


@dataclass
class File(Body):
    media_type: str = "multipart/form-data"
    alias: Optional[str] = None



import typing
from typing import Any, Awaitable, Callable, Coroutine, MutableMapping, TypeVar

Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]

ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]

DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., typing.Any])

AnyFunc = Callable[..., Coroutine[Any, Any, Any]]



import logging
import time
from pathlib import Path
import random
from typing import Dict, Optional

import yaml
from github import Github
from pydantic import BaseModel, BaseSettings, SecretStr

awaiting_label = "awaiting review"
lang_all_label = "lang-all"
approved_label = "approved-2"
translations_path = Path(__file__).parent / "translations.yml"


class Settings(BaseSettings):
    github_repository: str
    input_token: SecretStr
    github_event_path: Path
    github_event_name: Optional[str] = None
    input_debug: Optional[bool] = False


class PartialGitHubEventIssue(BaseModel):
    number: int


class PartialGitHubEvent(BaseModel):
    pull_request: PartialGitHubEventIssue


if __name__ == "__main__":
    settings = Settings()
    if settings.input_debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    logging.debug(f"Using config: {settings.json()}")
    g = Github(settings.input_token.get_secret_value())
    repo = g.get_repo(settings.github_repository)
    if not settings.github_event_path.is_file():
        raise RuntimeError(
            f"No github event file available at: {settings.github_event_path}"
        )
    contents = settings.github_event_path.read_text()
    github_event = PartialGitHubEvent.parse_raw(contents)
    translations_map: Dict[str, int] = yaml.safe_load(translations_path.read_text())
    logging.debug(f"Using translations map: {translations_map}")
    sleep_time = random.random() * 10  # random number between 0 and 10 seconds
    pr = repo.get_pull(github_event.pull_request.number)
    logging.debug(
        f"Processing PR: {pr.numeric}, with anti-race condition sleep time: {sleep_time}"
    )
    if pr.state == "open":
        logging.debug(f"PR is open: {pr.numeric}")
        label_strs = set([label.name for label in pr.get_labels()])
        if lang_all_label in label_strs and awaiting_label in label_strs:
            logging.info(
                f"This PR seems to be a language translation and awaiting reviews: {pr.numeric}"
            )
            if approved_label in label_strs:
                message = (
                    f"It seems this PR already has the approved label: {pr.numeric}"
                )
                logging.error(message)
                raise RuntimeError(message)
            langs = []
            for label in label_strs:
                if label.startswith("lang-") and not label == lang_all_label:
                    langs.append(label[5:])
            for lang in langs:
                if lang in translations_map:
                    num = translations_map[lang]
                    logging.info(
                        f"Found a translation issue for language: {lang} in issue: {num}"
                    )
                    issue = repo.get_issue(num)
                    message = f"Good news everyone! 😉 There's a new translation PR to be reviewed: #{pr.numeric} 🎉"
                    already_notified = False
                    time.sleep(sleep_time)
                    logging.info(
                        f"Sleeping for {sleep_time} seconds to avoid race conditions and multiple comments"
                    )
                    logging.info(
                        f"Checking current comments in issue: {num} to see if already notified about this PR: {pr.numeric}"
                    )
                    for comment in issue.get_comments():
                        if message in comment.body:
                            already_notified = True
                    if not already_notified:
                        logging.info(
                            f"Writing comment in issue: {num} about PR: {pr.numeric}"
                        )
                        issue.create_comment(message)
                    else:
                        logging.info(
                            f"Issue: {num} was already notified of PR: {pr.numeric}"
                        )
    else:
        logging.info(
            f"Changing labels in a closed PR doesn't trigger comments, PR: {pr.numeric}"
        )
    logging.info("Finished")



from squall.exceptions import (
    RequestHeadValidationError,
    RequestPayloadValidationError,
    ResponsePayloadValidationError,
)
from squall.requests import Request
from squall.responses import JSONResponse, PrettyJSONResponse
from starlette.exceptions import HTTPException
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)


async def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    headers = getattr(exc, "headers", None)
    if headers:
        return JSONResponse(
            {"details": exc.detail}, status_code=exc.status_code, headers=headers
        )
    else:
        return JSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def request_payload_validation_exception_handler(
    request: Request, exc: RequestPayloadValidationError
) -> PrettyJSONResponse:
    return PrettyJSONResponse(
        status_code=HTTP_422_UNPROCESSABLE_ENTITY,
        content={"details": exc.errors},
    )


async def request_head_validation_exception_handler(
    request: Request, exc: RequestHeadValidationError
) -> PrettyJSONResponse:
    return PrettyJSONResponse(
        status_code=HTTP_400_BAD_REQUEST,
        content={"details": exc.errors},
    )


async def response_payload_validation_exception_handler(
    request: Request, exc: ResponsePayloadValidationError
) -> PrettyJSONResponse:
    return PrettyJSONResponse(
        status_code=HTTP_500_INTERNAL_SERVER_ERROR,
        content={"details": exc.errors},
    )



from squall import Router, Squall
from squall.responses import HTMLResponse, JSONResponse, PlainTextResponse
from squall.testclient import TestClient


class OverrideResponse(JSONResponse):
    media_type = "application/x-override"


json_type = "application/json"
text_type = "text/plain; charset=utf-8"
html_type = "text/html; charset=utf-8"
override_type = "application/x-override"


def test_defaults():
    app = Squall()
    router = Router(prefix="/r")

    @router.get("/")
    def get_root():
        return {"msg": "Hello World"}

    app.include_router(router)
    client = TestClient(app)
    response = client.get("/r")
    assert response.json() == {"msg": "Hello World"}
    assert response.headers["content-type"] == json_type


def test_override():
    app = Squall()
    router = Router(prefix="/r", default_response_class=OverrideResponse)

    @router.get("/override")
    def get_default():
        return {"msg": "Hello World"}

    @router.get("/text", response_class=PlainTextResponse)
    def get_overridden():
        return "Hello World"

    app.include_router(router)
    client = TestClient(app)

    response = client.get("/r/override")
    assert response.json() == {"msg": "Hello World"}
    assert response.headers["content-type"] == override_type

    response = client.get("/r/text")
    assert response.text == "Hello World"
    assert response.headers["content-type"] == text_type


def test_nesting():
    app = Squall()
    router = Router(prefix="/r", default_response_class=OverrideResponse)
    nested_router = Router(prefix="/n", default_response_class=HTMLResponse)

    @nested_router.get("/html")
    def get_default_nested():
        return "Hello HTML"

    @router.get("/override")
    def get_default():
        return {"msg": "Hello World"}

    @router.get("/text", response_class=PlainTextResponse)
    def get_overridden():
        return "Hello World"

    router.include_router(nested_router)
    app.include_router(router)
    client = TestClient(app)

    response = client.get("/r/override")
    assert response.json() == {"msg": "Hello World"}
    assert response.headers["content-type"] == override_type

    response = client.get("/r/text")
    assert response.text == "Hello World"
    assert response.headers["content-type"] == text_type

    response = client.get("/r/n/html")
    assert response.text == "Hello HTML"
    assert response.headers["content-type"] == html_type



import logging
import sys
from pathlib import Path
from typing import Optional

import httpx
from github import Github
from github.PullRequest import PullRequest
from pydantic import BaseModel, BaseSettings, SecretStr, ValidationError

github_api = "https://api.github.com"


class Settings(BaseSettings):
    github_repository: str
    github_event_path: Path
    github_event_name: Optional[str] = None
    input_token: SecretStr
    input_deploy_url: str


class PartialGithubEventHeadCommit(BaseModel):
    id: str


class PartialGithubEventWorkflowRun(BaseModel):
    head_commit: PartialGithubEventHeadCommit


class PartialGithubEvent(BaseModel):
    workflow_run: PartialGithubEventWorkflowRun


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    settings = Settings()
    logging.info(f"Using config: {settings.json()}")
    g = Github(settings.input_token.get_secret_value())
    repo = g.get_repo(settings.github_repository)
    try:
        event = PartialGithubEvent.parse_file(settings.github_event_path)
    except ValidationError as e:
        logging.error(f"Error parsing event file: {e.errors()}")
        sys.exit(0)
    use_pr: Optional[PullRequest] = None
    for pr in repo.get_pulls():
        if pr.head.sha == event.workflow_run.head_commit.id:
            use_pr = pr
            break
    if not use_pr:
        logging.error(
            f"No PR found for hash: {event.workflow_run.head_commit.id}"
        )
        sys.exit(0)
    github_headers = {
        "Authorization": f"token {settings.input_token.get_secret_value()}"
    }
    url = f"{github_api}/repos/{settings.github_repository}/issues/{use_pr.numeric}/comments"
    logging.info(f"Using comments URL: {url}")
    response = httpx.post(
        url,
        headers=github_headers,
        json={
            "body": f"📝 Docs preview for commit {use_pr.head.sha} at: {settings.input_deploy_url}"
        },
    )
    if not (200 <= response.status_code <= 300):
        logging.error(f"Error posting comment: {response.text}")
        sys.exit(1)
    logging.info("Finished")



# from squall import Depends, Response, Squall
# from squall.testclient import TestClient
#
# app = Squall()
#
#
# async def response_status_setter(response: Response):
#     response.status_code = 201
#
#
# async def parent_dep(result=Depends(response_status_setter)):
#     return result
#
#
# @app.get("/", dependencies=[Depends(parent_dep)])
# async def get_main():
#     return {"msg": "Hello World"}
#
#
# client = TestClient(app)
#
#
# def test_dependency_set_status_code():
#     response = client.get("/")
#     assert response.status_code == 201, response.text
#     assert response.json() == {"msg": "Hello World"}



import inspect
import typing
from dataclasses import asdict, is_dataclass
from decimal import Decimal
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    get_args,
    get_origin,
)

from squall.bindings import RequestField
from squall.params import (
    Body,
    CommonParam,
    Cookie,
    File,
    Form,
    Header,
    Num,
    Path,
    Query,
    Str,
)
from squall.requests import Request
from squall.utils import get_types
from starlette.websockets import WebSocket


class HeadParam:
    def __init__(self, name: str, value: inspect.Parameter, source: str) -> None:
        self._annotation = value.annotation
        self._default = value.default
        self._empty = value.empty
        self.name = name
        self.source = source
        self.is_array, self.convertor = self.get_convertor()
        self.validate, self.statements = self.get_validation_statements()

    @property
    def alias(self) -> str:
        return getattr(self._default, "alias", None) or self.name

    @property
    def default(self) -> Any:
        default = Ellipsis
        if isinstance(self._default, CommonParam):
            default = self._default.default
        elif type(self._default) in (int, float, Decimal, str, bytes, type(None)):
            default = self._default
        return default

    def get_validation_statements(self) -> Tuple[Optional[str], Dict[str, Any]]:
        validate, statements = None, {}
        if hasattr(self._default, "valid") and isinstance(
            self._default.valid, (Num, Str)
        ):
            statements = asdict(self._default.valid)
            validate = self._default.valid.in_.value
        return validate, statements

    def get_convertor(self) -> Tuple[bool, str]:
        is_array, convertor = False, "str"
        args, alias = get_args(self._annotation), get_origin(self._annotation)
        if alias is Union:
            if type(None) in args:
                if get_origin(args[0]) == list:
                    is_array, _convertor = True, get_args(args[0])[0]
                else:
                    is_array, _convertor = False, args[0]
                convertor = getattr(_convertor, "__name__", None)
        elif alias == list:
            try:
                is_array, convertor = True, args[0].__name__
            except Exception:
                """List[Any] should be implemented"""
        elif (
            hasattr(self._annotation, "__name__")
            and self._annotation.__name__ != "_empty"
        ):
            convertor = self._annotation.__name__
        else:
            assert not alias, f"Convertor for {self.name} unknown"
        return is_array, convertor


def get_handler_head_params(func: Callable[..., Any]) -> List[HeadParam]:
    signature = inspect.signature(func)
    results = []
    for k, v in signature.parameters.items():
        source = None
        if isinstance(v.default, (Query, Path, Cookie, Header)):
            source = v.default.in_.value
        elif v.annotation == WebSocket:
            continue
        elif v.default is v.empty:
            is_model = is_valid_body_model(v.annotation)
            is_affilated = get_annotation_affiliation(v.annotation, v.default)
            if not (is_model or is_affilated):
                source = "path_params"
        elif type(v.default) in (int, float, Decimal, str, bytes, type(None)):
            source = "path_params"

        if source is not None:
            results.append(HeadParam(name=k, value=v, source=source))

    return results


def get_annotation_affiliation(annotation: Any, default: Any) -> Optional[Any]:
    """Helper for classifying affiliation of parameter

    :param annotation: annotation record
    :returns: classified value or None
    """
    args, alias = get_args(annotation), get_origin(annotation)
    # if alias and alias == list:
    annotation = args[0] if alias == list else annotation

    if annotation == Request:
        return "request"
    elif isinstance(default, (Form, File)):
        return "form"
    return None


def is_valid_body_model(annotation: Any) -> bool:
    """Checks if the annotation is a valid type and includes dataclass"""
    valid = {typing.Union, type(None), list, set, tuple}

    models = []
    for i in get_types(annotation):
        if i in valid:
            continue
        if is_dataclass(i):
            models.append(i)
        else:
            return False
    return len(models) == 1


def get_handler_body_params(func: Callable[..., Any]) -> List[Dict[str, Any]]:
    """Reads meta information from callable inspection.
    Filter out only parameters that appear from BODY.
    Provides all necessary data for build validation.

    :param func: callable for inspection
    """
    signature = inspect.signature(func)
    results = []
    for k, v in signature.parameters.items():
        param: Dict[str, Any] = {"name": k}
        annotation = v.annotation

        if annotation == Request:
            param["kind"] = "request"
        elif isinstance(v.default, (Form, File)):
            param["kind"] = "form"
            param["model_class"] = annotation
        elif isinstance(v.default, Body):
            param["kind"] = "body"
        else:
            continue

        results.append(param)

    return results


def get_handler_request_fields(func: Callable[..., Any]) -> List[RequestField]:
    """Returns all fields that match as request schema"""
    signature = inspect.signature(func)
    results = []
    for name, v in signature.parameters.items():
        if is_valid_body_model(v.annotation):
            settings = v.default if isinstance(v.default, Body) else None
            field = RequestField(name, model=v.annotation, settings=settings)
            results.append(field)
    return results
